// Source               ::  MALP_1Sheeld_V1

// Description          ::  Basic vehicle control program using 1Sheeld and Smartphone/Tablet 
//                                ...Accelerometer controls throttle, steering, and forward/backwards
//                                ...Slider acts as ignition switch and throttle governor
//
//                          This program will also ...
//                                ...use an ultrasonic sensor, if installed and code activated, to prevent it from running into obstacles.
//
//                          MALP "Mobile Analytic Laboratory Probe"
//                                http://stargate.wikia.com/wiki/Mobile_Analytic_Laboratory_Probe
//


//                          Recommended components ( or equivalent ) 

//                          1)   Dagu 4WD Thumper (with 2 Degrees of Freedom (2DoF) Robotic Arm, optional)
//                                     http://www.robotshop.com/dagu-4wd-thumper-mobile-platform-with-2dof-arm-black-751.html
//                          2)   Sabertooth Dual 25A 6V-24V Regenerative Motor Driver
//                                     http://www.robotshop.com/dimension-engineering-sabertooth-2x25.html
//                          3)   6V, 10000mAh Rechargeable NiMh Battery
//                                     http://www.robotshop.com/6v-10000mah-rechargeable-nimh-battery.html
//                          4)   1Sheeld
//                                     http://www.robotshop.com/1sheeld-android-smartphone-multi-purpose-shield-arduino.html
//                          5)   Arduino Uno
//                                     http://www.robotshop.com/arduino-uno-usb-microcontroller-rev-3.html
//                          6)   Ultrasonic Range Finder
//                                     http://www.robotshop.com/seeedstudio-ultrasonic-range-finder.html



// Original Program     ::  MALP control & program standards     - Joe M  / arduino (at) bellaliant.net   / March 2016
//                          

// Modified By          ::  Your Name Here / Contact / Date  



//  This program is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License 
//  as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  Lesser General Public License for more details.
//
//  For a copy of the GNU Lesser General Public License, 
//  write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA



// **** I M P O R T A N T **** **** I M P O R T A N T **** **** I M P O R T A N T **** **** I M P O R T A N T **** **** I M P O R T A N T **** **** I M P O R T A N T ****
// ***********************************************************************************************************************************************************************
// Pinouts... IMPORTANT!!!! ... set the pinouts to your hardware configuration
//
// This program uses very few components.  Constants are included for your own expansion.
//
// ***********************************************************************************************************************************************************************

// ********** Digital Pins D0 thru D13 **********

// Using 0 and 1 disables upload!     

#define DO_NOT_USE_0        0
#define DO_NOT_USE_1        1

#define NOT_USED_2          2              

//#define RANGEFINDER_PIN      3   // The ultrasonic rangefinder mounted on the front of the MALP (if installed)
                                   // Comment out this define if rangefinder not installed 

#define NOT_USED_4          4        
#define NOT_USED_5          5
#define NOT_USED_6          6
#define NOT_USED_7          7
#define NOT_USED_8          8
#define NOT_USED_9          9
#define NOT_USED_10         10

 
#define SABER_RX_PIN        11  // Serial receive pin... NOT USED (but still init'd)
#define SABER_TX_PIN        12  // Serial transmit pin.  This pin is used to control the motors
#define NOT_USED_13         13  


// ********** Analog Pins A0 thru A5 **********

// FYI... using analog pins for digital I/O is just the same as using digital ones.

// A0 is referred to as Pin 14     A1 is referred to as Pin 15
// A2 is referred to as Pin 16     A3 is referred to as Pin 17
// A4 is referred to as Pin 18     A5 is referred to as Pin 19

#define NOT_USED_A0         A0  
#define NOT_USED_A1         A1 
#define NOT_USED_A2         A2 
#define NOT_USED_A3         A3 
#define NOT_USED_A4         A4  
#define NOT_USED_A5         A5       



// *********************************************************************************************************************************************************************
// Required includes <--------------------------------------------------------------------------------------------------------------------------------------------------
// *********************************************************************************************************************************************************************

// 1Sheeld
// Not including these constants will compile the entire 1Sheeld library into your sketch,
// making the compiled size larger than required
#define CUSTOM_SETTINGS
#define INCLUDE_ACCELEROMETER_SENSOR_SHIELD
#define INCLUDE_SLIDER_SHIELD
#include <OneSheeld.h>

// Required for communicating with Sabertooth 
#include <SoftwareSerial.h>


#ifdef RANGEFINDER_PIN
// Optional ultrasonic sensor library (if installed) 
#include <Ultrasonic.h>
#endif




// *********************************************************************************************************************************************************************
//  Global Variables.
//  Professional programmers cringe at the thought of globals, but they seem to be generally acceptable in the "wild west" Arduino world.
//  Prudent use of them simplifies program design since variable passing via parameters between functions are reduced.
// *********************************************************************************************************************************************************************

// An homage to every program I've written since my WATFIV days.
// Every program needs an "int i", right?
int i;


// Constants for Sabertooth 2x25 Motor Controller
// Set to 9600 through Sabertooth dip switches
#define SABER_BAUDRATE 9600


// Global variables for Motor Control

SoftwareSerial SaberSerial = SoftwareSerial( SABER_RX_PIN, SABER_TX_PIN );


static boolean actionRequired;

// Enumerated data types help make code more readable
// http://playground.arduino.cc/Code/Enum

enum ThrottleAndParkReverseDriveSelector { 
  
  REVERSE_4x4 , REVERSE_3x4 , REVERSE_2x4 , REVERSE_3x8 , REVERSE_1x4 ,
  STOPNOW_0x4 ,
  FORWARD_1x4 , FORWARD_3x8 , FORWARD_2x4 , FORWARD_3x4 , FORWARD_4x4
  
} throttle = STOPNOW_0x4 , prevThrottle = STOPNOW_0x4;

enum SteeringWheel { 
  
  LEFT , STRAIGHT , RIGHT
   
} steering = STRAIGHT , prevSteering = STRAIGHT;



// Constants for Right , Left Motor Control via Sabertooth Mode 3 Simplified Serial Mode.
// Full , 3/4 , Half, 1/4 Speeds , plus Stop , provided.
// If you want to add other speed levels, just follow the provided scheme.

#define FORWARD_4x4_R_MOTOR   255
#define FORWARD_3x4_R_MOTOR   240
#define FORWARD_2x4_R_MOTOR   224
#define FORWARD_3x8_R_MOTOR   216  
#define FORWARD_1x4_R_MOTOR   208
#define STOPNOW_0x4_R_MOTOR   192
#define REVERSE_1x4_R_MOTOR   175
#define REVERSE_3x8_R_MOTOR   167
#define REVERSE_2x4_R_MOTOR   159
#define REVERSE_3x4_R_MOTOR   143
#define REVERSE_4x4_R_MOTOR   128

#define FORWARD_4x4_L_MOTOR   127
#define FORWARD_3x4_L_MOTOR   112
#define FORWARD_2x4_L_MOTOR    96
#define FORWARD_3x8_L_MOTOR    88
#define FORWARD_1x4_L_MOTOR    80
#define STOPNOW_0x4_L_MOTOR    64
#define REVERSE_1x4_L_MOTOR    48
#define REVERSE_3x8_L_MOTOR    40
#define REVERSE_2x4_L_MOTOR    32
#define REVERSE_3x4_L_MOTOR    16
#define REVERSE_4x4_L_MOTOR     1

#define STOPALL_0x0_B_MOTOR    (byte)0x00   // Stop both motors



#ifdef RANGEFINDER_PIN     // "ifdef's" are one of those obscure C++ features most beginning programmers don't know about.
                           // It is a neat feature that allows "conditional compilation".  In the case of this program,
                           // the various code segments using the ultrasonic sensor will only be included in the compiled executable
                           // if you defined RANGEFINDER_PIN above ( ie #define RANGEFINDER_PIN 3 ).
                           // For more on "conditional compilation", go to http://www.cplusplus.com/doc/tutorial/preprocessor/

// Constants and variables for the ultrasonic sensor (if installed)
#define   DANGER_CLOSE    20  // Stop the MALP within this distance in cms from obstacle in front.


#endif





// **********************************************************************************************************************************************************************
// Function Prototypes <-------------------------------------------------------------------------------------------------------------------------------------------------                                                                                                                                               
// **********************************************************************************************************************************************************************

boolean   getCurrentThrottleSteeringSettings  ( void );
void      malpAction                          ( void );

#ifdef RANGEFINDER_PIN
boolean   obstacleTooClose                    ( void );
#endif


   
// *********************************************************************************************************************************************************************
// Main program <-------------------------------------------------------------------------------------------------------------------------------------------------------                                                                                                                                                                                                                   
// *********************************************************************************************************************************************************************
 
void setup( ) {
    
  // Connect to your Smartphone/Tablet for pilot control
  OneSheeld.begin( );
  
  // Transmit pin for Sabertooth communication
  pinMode( SABER_TX_PIN, OUTPUT );
  
  // Connect to the Sabertooth for drive motor control
  SaberSerial.begin( SABER_BAUDRATE ); 

  // Wait for the Sabertooth to initialize
  delay( 2000 );
  
  // TEST TEST TEST TEST
  // Only activate this code to test your motors.
  // This function never terminates, so if you call it,
  // this sketch will do nothing else but spin the motors forward and back.
  // motorTest( );

    
  // Ensure MALP is "parked"
  throttle = STOPNOW_0x4;
  steering = STRAIGHT;
  malpAction( );


  #ifdef RANGEFINDER_PIN 
  // If Ultrasonic sensor installed...     
  pinMode( RANGEFINDER_PIN, INPUT );       
  #endif


}





void loop( ) {

  actionRequired = getCurrentThrottleSteeringSettings ( );
  
  #ifdef RANGEFINDER_PIN
  // Override pilot command if obstacle detected.
  // If no obstacle, keep pilot command (if any).
  actionRequired = ( obstacleTooClose ( ) ) ? true : actionRequired;
  #endif

  if ( actionRequired ) malpAction( ); 
  
  
}




// *********************************************************************************************************************************************************************
// Functions <---------------------------------------------------------------------------------------------------------------------------------------------------------- 
// *********************************************************************************************************************************************************************


boolean getCurrentThrottleSteeringSettings  ( void ) {

  // Incoming parameters :   Nothing.
    
  // Global variables    :   Uses    -   
  //                         Changes -  prevThrottle , throttle
  
  // Responsibilities    :   Reads Smartphone/Tablet accelerometer via 1Sheeld,
  //                         and maps Z and X readings to throttle and steering commands
  //                         to be executed in malpAction(). 

  //                         If a throttle or steering command is the same as last reading,
  //                         this function reports "no new command detected".

  //                         1Sheeld "Slider" used to determine ignition status and govern throttle
  
  // Returns             :   true if new pilot command detected, false otherwise

  
  // Make local variables static to save the CPU from having to
  // create them every time this function is called
  static boolean  commandDetected = false; 
  static int newThrottleInput;
  static int newSteeringInput;
  
  // Save the previous throttle/steering values before reading the new ones
  prevThrottle = throttle;
  prevSteering = steering;
 
  // Accelerometer X Axis controls steering 
  //    ... left / right
  newSteeringInput = ( (int) AccelerometerSensor.getX() );

  if ( newSteeringInput >= 3 )        steering = LEFT;
  else if ( newSteeringInput <= -3 )  steering = RIGHT; 
  else                                steering = STRAIGHT;

    
  // Accelerometer Z Axis controls throttle 
  //    ... the more you tilt the accelerometer forward, the faster the MALP will go forward
  //    ... the more you tilt the accelerometer back, the faster the MALP will go in reverse
  newThrottleInput = ( (int) AccelerometerSensor.getZ() ) / 2 ;  // Increase divisor to decrease accelerometer sensitivity

  
  // The Slider is used to restrict the throttle to a single low speed 
  if( Slider.getValue() < 128 ) {
      if ( newThrottleInput > 0 )        newThrottleInput =  99;
      else if ( newThrottleInput < 0 )   newThrottleInput = -99;
  }
  
  
  switch ( newThrottleInput ) {  
                                  
    case -99:  throttle = REVERSE_3x8;   break;     // Throttle is restricted
    
    case  -3:  throttle = REVERSE_3x4;   break;

    case  -2:  throttle = REVERSE_2x4;   break;
    
    case  -1:  throttle = REVERSE_1x4;   break;
    
    case   0:  throttle = STOPNOW_0x4;   break;

    case   1:  throttle = FORWARD_1x4;   break;

    case   2:  throttle = FORWARD_2x4;   break;

    case   3:  throttle = FORWARD_3x4;   break;

    case  99:  throttle = FORWARD_3x8;   break;   // Throttle is restricted

    // Anything else, full speed...
    default :  throttle = ( newThrottleInput > 0 ) ? FORWARD_4x4 : REVERSE_4x4;
               break;

  }

  
  // If "ignition" is not turned on (via Slider), override any pilot commands and stop MALP
  if( Slider.getValue() < 20 ) throttle = STOPNOW_0x4;

  commandDetected = (  ( throttle != prevThrottle ) || ( steering != prevSteering ) ) ? true : false ;
  
  return commandDetected;

}


void malpAction ( void ) {

  // Incoming parameters :   Nothing.
    
  // Global variables    :   Uses    -  throttle, steering
  //                         Changes -  
  
  // Responsibilities    :   Alters motor speed/direction depending on previously
  //                         calculated throttle and steering settings
  
  // Returns             :   Nothing
  
   
  // Communicate with Sabertooth using Mode 3 Simplified Serial Mode (ssm)... 
  // ... a single byte of data represents the desired speed/direction for a motor
  static byte ssm_L_Motor = 0 , ssm_R_Motor = 0;
  
  // Apply current throttle and steering settings
  switch ( throttle ) {

    case REVERSE_4x4:
    
        ssm_L_Motor = REVERSE_4x4_L_MOTOR;
        ssm_R_Motor = REVERSE_4x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_4x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_4x4_R_MOTOR;  break;
        }
        
        break;
     
    case REVERSE_3x4:
    
        ssm_L_Motor = REVERSE_3x4_L_MOTOR;
        ssm_R_Motor = REVERSE_3x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_3x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_3x4_R_MOTOR;  break;
        }

        break;       

    case REVERSE_2x4:
    
        ssm_L_Motor = REVERSE_2x4_L_MOTOR;
        ssm_R_Motor = REVERSE_2x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_2x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_2x4_R_MOTOR;  break;
        }
        
        break;    

    case REVERSE_3x8:
    
        ssm_L_Motor = REVERSE_3x8_L_MOTOR;
        ssm_R_Motor = REVERSE_3x8_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_3x8_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_3x8_R_MOTOR;  break;
        }
        
        break;

    case REVERSE_1x4:
    
        ssm_L_Motor = REVERSE_1x4_L_MOTOR;
        ssm_R_Motor = REVERSE_1x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_1x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_1x4_R_MOTOR;  break;
        }
        
        break;
 
    case STOPNOW_0x4:
              
        ssm_L_Motor = STOPNOW_0x4_L_MOTOR;
        ssm_R_Motor = STOPNOW_0x4_R_MOTOR;
        break;

    case FORWARD_1x4:
    
        ssm_L_Motor = FORWARD_1x4_L_MOTOR;
        ssm_R_Motor = FORWARD_1x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = REVERSE_1x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = REVERSE_1x4_R_MOTOR;  break;
        }
        
        break;

    case FORWARD_3x8:
    
        ssm_L_Motor = FORWARD_3x8_L_MOTOR;
        ssm_R_Motor = FORWARD_3x8_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = REVERSE_3x8_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = REVERSE_3x8_R_MOTOR;  break;
        }
        
        break;

    case FORWARD_2x4:
    
        ssm_L_Motor = FORWARD_2x4_L_MOTOR;
        ssm_R_Motor = FORWARD_2x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = REVERSE_2x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = REVERSE_2x4_R_MOTOR;  break;
        }
        
        break;

    case FORWARD_3x4:
    
        ssm_L_Motor = FORWARD_3x4_L_MOTOR;
        ssm_R_Motor = FORWARD_3x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_3x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_3x4_R_MOTOR;  break;
        }
        
        break;

    case FORWARD_4x4:
    
        ssm_L_Motor = FORWARD_4x4_L_MOTOR;
        ssm_R_Motor = FORWARD_4x4_R_MOTOR;

        switch ( steering ) {
          case LEFT:  ssm_L_Motor = FORWARD_4x4_L_MOTOR;  break;
          case RIGHT: ssm_R_Motor = FORWARD_4x4_R_MOTOR;  break;
        }
        
        break;

    default:
              
        break;
  }
 
  SaberSerial.write( ssm_R_Motor );
  SaberSerial.write( ssm_L_Motor ); 
  
}

#ifdef RANGEFINDER_PIN
boolean obstacleTooClose( void ) {
  
  // Incoming parameters :   Nothing.
    
  // Global variables    :   Uses    -   
  //                         Changes -  throttle
  
  // Responsibilities    :   Uses Ultrasonic sensor to detect obstacle in front.
  //                         Will set throttle to STOPNOW_0x4 if within DANGER_CLOSE range
  //                         (Optionally) Will set throttle to half speed if within 2.0 x DANGER_CLOSE range
  //                                      Will set throttle to quarter speed if within 1.5 x DANGER_CLOSE range
  
  // Returns             :   true if "danger close", false otherwise.
  


  // If the MALP isn't moving forward, there's no need to do anything here
  if ( throttle <= STOPNOW_0x4 ) return false;

  // Ultrasonic sensor object 
  static Ultrasonic ultrasonicFront(RANGEFINDER_PIN);

  // This will contain the calculated distance to obstacle
  static int frontRangeInCentimeters;
  
  ultrasonicFront.DistanceMeasure();                                                 // Get the current signal time
  frontRangeInCentimeters =   ultrasonicFront.microsecondsToCentimeters();           // Convert the time to centimeters
  
  // Possible obstacle... but because the ultrasonic is so unreliable, take a second look
  if ( frontRangeInCentimeters <= ( DANGER_CLOSE * 2 ) ) {
      
    ultrasonicFront.DistanceMeasure();                                                // Get the current signal time
    frontRangeInCentimeters += ultrasonicFront.microsecondsToCentimeters();           // Convert the time to centimeters
        
    frontRangeInCentimeters /= 2;   // Average the 2 readings
    
  }


  // Too close, stop!
  if ( frontRangeInCentimeters <= DANGER_CLOSE ) {
    
    throttle = STOPNOW_0x4;       
    return true;
        
  }
  
  // Getting really close to potential threat, slow down 
  if ( frontRangeInCentimeters <= ( DANGER_CLOSE * 1.5 ) ) {
    
    if ( throttle > FORWARD_1x4 ) { 
      throttle = FORWARD_1x4;      
      return true;
    }   
    
  }

  // Getting close to potential threat, slow down
  if ( frontRangeInCentimeters <= ( DANGER_CLOSE * 2.0 ) ) {
    
    if ( throttle > FORWARD_2x4 ) { 
      throttle = FORWARD_2x4;
      return true;
    }   
    
  }
  
  return false;
 
 }
#endif


void motorTest ( void ) {
  
  // Incoming parameters :   Nothing.
    
  // Global variables    :   Uses    -   
  //                         Changes -  
  
  // Responsibilities    :   Use this to test your motors.
  //                         If your motors do not perform to expectations, it may indicate...
  //                              1) Your battery/motor/controller configuration does not support that speed
  //                                 (especially the lower speeds, motors stall out due to lack of current)
  //                              2) Batteries are too weak... ensure you have a full charge
  
  // Returns             :   Nothing.  This function never terminates.

    int desiredTime = 2000;
    
    // You could disable the loop to make this a "one time" test
    while ( true ) {
      
      SaberSerial.write( FORWARD_1x4_R_MOTOR );  
      SaberSerial.write( FORWARD_1x4_L_MOTOR );
      delay ( desiredTime ) ;
      
      SaberSerial.write( FORWARD_2x4_R_MOTOR );  
      SaberSerial.write( FORWARD_2x4_L_MOTOR );
      delay ( desiredTime ) ;
      
      SaberSerial.write( FORWARD_3x4_R_MOTOR );  
      SaberSerial.write( FORWARD_3x4_L_MOTOR );
      delay ( desiredTime ) ;
            
      SaberSerial.write( FORWARD_4x4_L_MOTOR );  
      SaberSerial.write( FORWARD_4x4_R_MOTOR );
      delay ( desiredTime ) ;
      
      SaberSerial.write( STOPALL_0x0_B_MOTOR );
      delay ( desiredTime ) ;
      
      SaberSerial.write( REVERSE_1x4_R_MOTOR );   
      SaberSerial.write( REVERSE_1x4_L_MOTOR );
      delay ( desiredTime ) ;

      SaberSerial.write( REVERSE_2x4_R_MOTOR );   
      SaberSerial.write( REVERSE_2x4_L_MOTOR );
      delay ( desiredTime ) ;

      SaberSerial.write( REVERSE_3x4_R_MOTOR );   
      SaberSerial.write( REVERSE_3x4_L_MOTOR );
      delay ( desiredTime ) ;
           
      SaberSerial.write( REVERSE_4x4_R_MOTOR );   
      SaberSerial.write( REVERSE_4x4_L_MOTOR );
      delay ( desiredTime ) ;  
      
      SaberSerial.write( STOPALL_0x0_B_MOTOR );
      delay ( desiredTime ) ;
      
    }
  
}

